import csv
import os
import shutil
import tempfile
import zipfile
import copy

# Demjson required. "pip install demjson"
import demjson


def main(name_ext):
    """
    Open a LiPD file and extract JSON and CSV as one dictionary (per LiPD file)
    :param name_ext: (str) LiPD filename
    :return: (dict) Metadata dictionary
    """
    # Running as standalone.
    # os.chdir("PATH_HERE")
    # name_ext = "FILENAME.lpd"

    # Mark our starting directory
    dir_root = os.getcwd()

    # Record name w/o extension
    name = os.path.splitext(name_ext)[0]

    # Create tmp folder
    dir_tmp = tempfile.mkdtemp()

    # Unzip LiPD contents to tmp folder, then move there
    _unzip(name_ext, dir_tmp)
    dir_data = os.path.join(dir_tmp, name, 'data')
    os.chdir(dir_data)

    # Read JSON in as dictionary, then switch to new structure
    j = _read_json_from_file(os.path.join(dir_data, name + '.jsonld'))
    metadata = _old_to_new_structure(j)

    # Read CSV, and insert into dictionary. Move back to root
    paleodata_w_csv, csv_only = _add_csv_to_json(metadata['paleoData'])
    metadata['paleoData'] = paleodata_w_csv
    os.chdir(dir_root)

    # Trash tmp folder
    shutil.rmtree(dir_tmp)

    # Return dictionary
    return metadata


def batch():
    """
    NOTE: ALTERNATE TO MAIN(). BATCH PROCESS INSTEAD OF SINGLE PROCESS.
    Get list of LiPD files from current working directory.
    :return: (dict) Dictionary of dictionaries. Key: LiPD filename, Val: Metadata dictionary
    """
    # os.chdir("PATH_HERE")
    metadata_dictionary = {}
    dir_root = os.getcwd()
    lipd_files = _list_files('lpd')

    for name_ext in lipd_files:
        # Record name w/o extension
        name = os.path.splitext(name_ext)[0]

        # Create tmp folder
        dir_tmp = tempfile.mkdtemp()

        # Unzip LiPD contents to tmp folder. Move there
        _unzip(name_ext, dir_tmp)
        dir_data = os.path.join(dir_tmp, name, 'data')
        os.chdir(dir_data)

        # Read JSON in as dictionary, then switch to new structure
        j = _read_json_from_file(os.path.join(dir_data, name + '.jsonld'))
        metadata = _old_to_new_structure(j)

        # Read CSV, and insert into dictionary. Move back to root
        paleodata_w_csv, csv_only = _add_csv_to_json(metadata['paleoData'])
        metadata['paleoData'] = paleodata_w_csv
        os.chdir(dir_root)

        # Trash tmp folder
        shutil.rmtree(dir_tmp)

        # Return dictionary
        metadata_dictionary[name] = metadata

    return metadata_dictionary


def _unzip(name_ext, dir_tmp):
    """
    Unzip .lpd file contents to tmp directory.
    :param name_ext: (str) Name of lpd file with extension
    :param dir_tmp: (str) Tmp folder to extract contents to
    :return: None
    """
    # Unzip contents to the tmp directory
    try:
        with zipfile.ZipFile(name_ext) as f:
            f.extractall(dir_tmp)
    except FileNotFoundError:
        shutil.rmtree(dir_tmp)
    return


def _read_json_from_file(filename):
    """
    Import the JSON data from target file.
    :param filename: (str) Target file
    :return: (dict) JSON data
    """
    d = {}
    try:
        # Load json into dictionary
        d = demjson.decode_file(filename)
    except FileNotFoundError:
        print("LiPD object: Load(). file not found")
    return d


def _add_csv_to_json(d):
    """
    NOTE: ONLY FOR NEW STRUCTURE
    Using the given paleoData dictionary from the JSON metadata, retrieve CSV data from CSV files, and insert the CSV
    data columns to their respective JSON paleoData table columns.
    :param d: (dict) PaleoData dictionary
    :return: (dict) Modified original dictionary (dict) CSV column data
    """
    d2 = {}
    # Loop through each table in paleoData
    for table_name, table_content in d.items():
        # Create CSV entry into dictionary that contains all columns.
        d2[table_content['filename']] = _read_csv_to_columns(table_content['filename'])
        # Start putting CSV data into corresponding JSON metadata columns under 'values' key.
        for col_name, col_content in table_content['columns'].items():
            col_content['values'] = d2[table_content['filename']][col_content['number']-1]
    return d, d2


def _read_csv_to_columns(filename):
    """
    Opens the target CSV file and creates a dictionary with one list for each CSV column.
    :param filename: (str) Filename
    :return: (dict) CSV data. Keys: Column number(int), Values: Column data (list)
    """
    d = {}
    try:
        with open(filename, 'r') as f:
            r = csv.reader(f, delimiter=',')
            # Create a dict with X lists corresponding to X columns
            for idx, col in enumerate(next(r)):
                d[idx] = []
            # Start iter through CSV data
            for row in r:
                for idx, col in enumerate(row):
                    # Append the cell to the correct column list
                    try:
                        d[idx].append(float(col))
                    except ValueError:
                        d[idx].append(col)
    except FileNotFoundError:
        print('CSV FileNotFound: ' + filename)
    return d


def _list_files(x):
    """
    Lists file(s) in given path of the X type.
    :param x: (str) File extension that we are interested in.
    :return: (list of str) File name(s) to be worked on
    """
    file_list = []
    for file in os.listdir():
        if file.endswith(x):
            file_list.append(file)
    return file_list


def _old_to_new_structure(d):
    """
    Restructure JSON to the new format. Table and columns referenced by name, not index. (dictionaries)
    :param d:(dict) JSON metadata
    :return: (dict) JSON metadata
    """
    # PaleoData dict
    tmp_p = {}

    for table in d['paleoData']:
        # All tables, all columns, table name
        tmp_t = {}
        tmp_c = {}
        t_name = table['paleoDataTableName']

        # paleoDataTableName and variableName are current
        # tableName and parameter are deprecated

        # Restructure columns into tmp_c
        for col in table['columns']:
            c_name = col['variableName']
            tmp_c[c_name] = col

        # Move table strings into tmp_t
        for k, v in table.items():
            if isinstance(v, str):
                tmp_t[k] = v

        # Move new column structure to table
        tmp_t['columns'] = copy.deepcopy(tmp_c)

        # Move table into tmp_p
        tmp_p[t_name] = copy.deepcopy(tmp_t)

    # Overwrite original paleoData dictionary with new dictionary
    d['paleoData'] = tmp_p
    return d


if '__name__' == '__main__':
    main()

main()



"""
GLOBAL LIST OF ALTERNATES AND SYNONYMS
"""

# Missing value name appears as many variations. Try to account for all of them
ALTS_MV = ['missing value', 'missing values', 'missingvalue', 'missingvalues', 'missing_values',
                    'missing variables', 'missing_variables', 'missingvariables']


from collections import OrderedDict
import os
import csv
import re
import copy

from ..helpers.jsons import write_json_to_file, EMPTY
from ..helpers.bag import *
from ..helpers.zips import *


# GLOBALS

# List of items that we don't want to output
IGNORE_KEYS = ['earliest_year', 'most_recent_year', 'note']
IGNORE_VAR_LINES = ['have no #', 'variables', 'lines begin with #', 'double marker', 'line variables format',
                    'line format', 'c or n for character or numeric', 'preceded by', "filename"]
IGNORE_DATA_LINES = ['have no #', 'tab-delimited text', 'age ensembles archived']
IGNORE_BLANKS = ['\n', '', ' ', '# ', '#']
# RE_BLANKS = re.compile(r'(#\s*(.)*)', re.S)
RE_VAR = re.compile(r'#{2}(\w)+')
RE_VAR_SPLIT = re.compile(r'(\w+)(\s+)([\w\W\s]+)')

# Missing value name appears as many variations. Try to account for all of them
MISSING_VAL_ALTS = ['missing value', 'missing values', 'missingvalue', 'missingvalues', 'missing_values',
                    'missing variables', 'missing_variables', 'missingvariables']

# Lists for what keys go in specific dictionary blocks
SITE_INFO = {'lat': ['northernmostlatitude', 'northernmost latitude', 'northernmost_latitude',
                     'southernmostlatitude', 'southernmost latitude', 'southernmost_latitude'],
             'lon': ['easternmostlongitude', 'easternmost longitude', 'easternmost_longitude',
                     'westernmostlongitude', 'westernmost longitude', 'westernmost_longitude'],
             'properties': ['location', 'country', 'elevation', 'site_name', 'region'],
             }
FUNDING_LIST = ['funding_agency_name', 'grant']

class NOAA_LPD(object):

    def __init__(self, dir_root, dir_tmp, name):
        self.dir_root = dir_root
        self.dir_tmp = dir_tmp
        self.name = name
        self.name_txt = name + '.txt'
        self.name_lpd = name + '.lpd'
        self.metadata = {}

    def main(self):
        """
        Convert a NOAA text file into a lipd file. CSV files will be created if chronology or data sections are available.
        :return: (dict) Metadata Dictionary
        """
        # Run the file through the parser
        # Sets self.metadata, Creates CSVs in dir_tmp
        self.__parse()
        os.chdir(self.dir_tmp)

        # Dump Metadata JSON to dir_tmp
        write_json_to_file(self.name + '.json', self.metadata)

        # Create a bagit bag
        new_bag = create_bag(self.dir_tmp)
        open_bag(self.dir_tmp)
        new_bag.save(manifests=True)

        # Zip the bag
        os.chdir(self.dir_root)
        re_zip(self.dir_tmp, self.name, self.name_lpd)
        os.rename(self.name_lpd + '.zip', self.name_lpd)

        return

    def __parse(self):
        """
        Main parser
        Accept the text file. We'll open it, read it, and return a compiled dictionary to write to a json file
        May write a chronology CSV and a data CSV if those sections are available
        :param file (str):
        :param path (str):
        :param filename (str):
        :return:
        """
        # Strings
        last_insert = None
        missing_str = ''

        # Counters
        grant_id = 0
        funding_id = 0
        data_col_ct = 1
        line_num = 0

        # Boolean markers
        description_on = False
        publication_on = False
        abstract_on = False
        site_info_on = False
        chronology_on = False
        chron_vars_start = True
        variables_on = False
        data_vals_on = False
        data_on = False

        # Lists
        lat = []
        lon = []
        pub = []
        funding = []
        temp_abstract = []
        temp_description = []
        chron_col_list = []
        data_var_names = []
        data_col_list = []
        data_tables = []

        # All dictionaries needed to create JSON structure
        temp_funding = OrderedDict()
        temp_pub = OrderedDict()
        vars_dict = OrderedDict()
        core_len = OrderedDict()
        geo_properties = OrderedDict()
        chron_dict = OrderedDict()
        data_dict_upper = OrderedDict()
        data_dict_lower = OrderedDict()
        final_dict = OrderedDict()

        # Open the text file in read mode. We'll read one line at a time until EOF
        with open(self.name_txt, 'r') as f:
            for line in iter(f):
                line_num += 1

                # PUBLICATION
                # There can be multiple publications. Create a dictionary for each one.
                if publication_on:

                    # End of the section. Add the dictionary for this one publication to the overall list
                    if '-----' in line:
                        temp_pub['abstract'] = ''.join(temp_abstract)
                        pub.append(temp_pub.copy())
                        temp_abstract.clear()
                        temp_pub.clear()
                        abstract_on = False
                        publication_on = False

                    elif abstract_on:
                        temp_abstract.append(self.__str_cleanup(line))

                    # Add all info into the current publication dictionary
                    else:
                        line = self.__str_cleanup(line)
                        key, value = self.__slice_key_val(line)
                        temp_pub[self.__camel_case(key)] = value
                        if key == 'Abstract':
                            abstract_on = True
                            temp_abstract.append(value)

                # DESCRIPTION
                # Descriptions are often long paragraphs spanning multiple lines, but don't follow the key/value format
                elif description_on:

                    # End of the section. Turn marker off and combine all the lines in the section
                    if '-------' in line:
                        description_on = False
                        value = ''.join(temp_description)
                        final_dict['description'] = value

                    # The first line in the section. Split into key, value
                    elif 'Description:' in line:
                        key, val = self.__slice_key_val(line)
                        temp_description.append(val)

                    # Keep a running list of all lines in the section
                    else:
                        line = self.__str_cleanup(line)
                        temp_description.append(line)

                # SITE INFORMATION (Geo)
                elif site_info_on:

                    if '-------' in line:
                        site_info_on = False

                    else:
                        line = self.__str_cleanup(line)
                        key, value = self.__slice_key_val(line)
                        if key.lower() in SITE_INFO['lat']:
                            lat.append(self.__convert_num(value))

                        elif key.lower() in SITE_INFO['lon']:
                            lon.append(self.__convert_num(value))

                        elif key.lower() in SITE_INFO['properties']:
                            if key.lower() == 'elevation':
                                val, unit = self.__split_name_unit(value)
                                geo_properties['elevation'] = {'value': self.__convert_num(val), 'unit': unit}
                            else:
                                geo_properties[self.__camel_case(key)] = value

                # CHRONOLOGY
                elif chronology_on:

                    # When reaching the end of the chron section, set the marker to off and close the CSV file
                    if '-------' in line:
                        chronology_on = False

                        # If nothing between the chronology start and the end barrier, then there won't be a CSV
                        if chron_start_line != line_num-1:
                            chron_csv.close()

                    # Special case for first line in chron section. Grab variables and open a new CSV file
                    elif chron_vars_start:
                        chron_vars_start = False

                        # Open CSV for writing
                        chron_filename = self.name + '-chronology.csv'
                        csv_path = os.path.join(self.dir_tmp, chron_filename)
                        chron_csv = open(csv_path, 'w', newline='')
                        cw = csv.writer(chron_csv)

                        # Split the line into a list of variables
                        chron_col_ct = 1
                        line = line.lstrip()
                        variables = line.split('|')

                        # Create a dictionary of info for each column
                        for index, var in enumerate(variables):
                            temp_dict = OrderedDict()
                            temp_dict['column'] = chron_col_ct
                            name, unit = self.__split_name_unit(var.replace('\n', '').lstrip().rstrip())
                            temp_dict['parameter'] = name
                            temp_dict['units'] = unit
                            chron_col_list.append(temp_dict)
                            chron_col_ct += 1
                        chron_dict['filename'] = chron_filename
                        chron_dict['chronTableName'] = 'Chronology'
                        chron_dict['columns'] = chron_col_list

                    # Split the line of data values, then write to CSV file
                    else:
                        values = line.split()
                        cw.writerow(values)

                # VARIABLES
                # Variables are the only lines that have a double # in front
                elif variables_on:

                    process_line = False

                    # End of the section. Turn marker off
                    if "------" in line:
                        variables_on = False
                    for item in IGNORE_VAR_LINES:
                        if item.lower() in line.lower():
                            process_line = False
                    for item in IGNORE_BLANKS:
                        if item == line:
                            process_line = False
                    m = re.match(RE_VAR, line)
                    if m:
                        process_line = True

                    # If the line isn't in the ignore list, then it's a variable line
                    if process_line:
                        # Split the line items, and cleanup
                        cleaned_line = self.__separate_data_vars(line)

                        # Add the items into a column dictionary
                        data_col_dict = self.__create_var_col(cleaned_line, data_col_ct)

                        # Keep a list of all variable names
                        data_var_names.append(data_col_dict['parameter'])

                        # Add the column dictionary into a final dictionary
                        data_col_list.append(data_col_dict)
                        data_col_ct += 1

                # DATA
                # Missing Value, Create data columns, and output Data CSV
                elif data_on:

                    # Do not process lines that are blank, template lines, or missing value
                    process_line = True

                    for item in IGNORE_DATA_LINES:
                        if item in line:
                            process_line = False
                    for item in IGNORE_BLANKS:
                        if item == line:
                            process_line = False
                    for item in MISSING_VAL_ALTS:
                        if item in line.lower():
                            process_line = False
                            line = self.__str_cleanup(line)
                            key, missing_str = self.__slice_key_val(line)

                    if process_line:
                        # Split the line at each space (There's one space between each data item)
                        values = line.split()

                        # Write all data values to CSV
                        if data_vals_on:
                            dw.writerow(values)

                        # Check for the line of variables
                        else:

                            var = self.__str_cleanup(values[0].lstrip())
                            # Check if a variable name is in the current line
                            if var.lower() == data_var_names[0].lower():
                                data_vals_on = True

                                # Open CSV for writing
                                data_filename = self.name + '-data.csv'
                                csv_path = os.path.join(self.dir_tmp, data_filename)
                                data_csv = open(csv_path, 'w', newline='')
                                dw = csv.writer(data_csv)

                # METADATA
                else:
                    # Line Continuation: Sometimes there are items that span a few lines.
                    # If this happens, we want to combine them all properly into one entry.
                    if '#' not in line and line not in IGNORE_BLANKS and last_insert is not None:
                        old_val = last_insert[old_key]
                        last_insert[old_key] = old_val + line

                    # No Line Continuation: This is the start or a new entry
                    else:
                        line = self.__str_cleanup(line)

                        # Grab the key and value from the current line
                        try:
                            # Split the line into key, value pieces
                            key, value = self.__slice_key_val(line)
                            lkey = key.lower()

                            # If there is no value, then we are at a section header.
                            # Data often has a blank value, so that is a special check.
                            if value is None or lkey == 'data':

                                # Turn on markers if we run into section headers
                                if lkey == 'description_and_notes':
                                    description_on = True
                                elif lkey == 'publication':
                                    publication_on = True
                                elif lkey == 'site_information':
                                    site_info_on = True
                                elif lkey == 'chronology':
                                    chronology_on = True
                                    chron_start_line = line_num
                                elif lkey == 'variables':
                                    variables_on = True
                                elif lkey == 'data':
                                    data_on = True

                            # For all
                            else:

                                # Ignore any entries that are specified in the skip list
                                if lkey not in IGNORE_KEYS:

                                    if lkey == 'core_length':
                                        val, unit = self.__split_name_unit(value)
                                        core_len['value'] = self.__convert_num(val)
                                        core_len['unit'] = unit
                                        last_insert = core_len

                                    # There can be multiple funding agencies and grants. Keep a list of dict entries
                                    elif lkey in FUNDING_LIST:
                                        if lkey == 'funding_agency_name':
                                            funding_id += 1
                                            key = 'agency'
                                        elif lkey == 'grant':
                                            grant_id += 1
                                            key = 'grant'
                                        temp_funding[key] = value

                                        # If both counters are matching, we are ready to add content to the funding list
                                        if grant_id == funding_id:
                                            funding.append(temp_funding.copy())
                                            temp_funding.clear()

                                    else:
                                        final_dict[self.__camel_case(key)] = value
                                        last_insert = final_dict

                                    # Keep track of old key in case we have a line continuation
                                    old_key = key

                        # Ignore any errors from NoneTypes that are returned from slice_key_val
                        except TypeError:
                            pass

        # Wait to close the data CSV until we reached the end of the text file
        try:
            data_csv.close()
        except NameError:
            print("Couldn't close Data CSV")

        # Piece together measurements block
        data_dict_upper['filename'] = data_filename
        data_dict_upper['paleoDataTableName'] = 'Data'
        data_dict_upper['missingValue'] = missing_str
        data_dict_upper['columns'] = data_col_list
        data_tables.append(data_dict_upper)

        # Piece together geo block
        geo = self.__create_coordinates(lat, lon)
        geo['properties'] = geo_properties

        # Piece together final dictionary
        final_dict['pub'] = pub
        final_dict['funding'] = funding
        final_dict['geo'] = geo
        final_dict['coreLength'] = core_len
        final_dict['chronology'] = chron_dict
        final_dict['paleoData'] = data_tables

        # Insert the data dictionaries into the final dictionary
        for k, v in vars_dict.items():
            data_dict_lower[k] = v

        # Set final_dict to self.
        self.metadata = final_dict

        return

    @staticmethod
    def __create_var_col(l, col_count):
        """
        Receive split list from separate_data_vars, and turn it into a dictionary for that column
        :param l (list):
        :param col_count (int):
        :return: (dict) d
        """
        # Format: what, material, error, units, seasonality, archive, detail, method,
        # C or N for Character or Numeric data, direction of relation to climate (positive or negative)
        d = OrderedDict()
        d['column'] = col_count
        for index, item in enumerate(l):
            if index == 0:
                d['parameter'] = item
            elif index == 1:
                d['description'] = item
            elif index == 2:
                d['material'] = item
            elif index == 3:
                d['error'] = item
            elif index == 4:
                d['units'] = item
            elif index == 5:
                d['seasonality'] = item
            elif index == 6:
                d['archive'] = item
            elif index == 7:
                d['detail'] = item
            elif index == 8:
                d['method'] = item
            elif index == 9:
                d['dataType'] = item
            elif index == 10:
                d['direction'] = item
        return d

    @staticmethod
    def __separate_data_vars(line):
        """
        For the variables section, clean up the line and return a list of each of the 10 items
        :param line (str):
        :return: (str)
        """
        combine = []
        if '#' in line:
            line = line.replace("#", "")
            line = line.lstrip()
        if line not in IGNORE_BLANKS and line not in EMPTY:

            m = re.match(RE_VAR_SPLIT, line)
            if m:
                combine.append(m.group(1))
                attr = m.group(3).split(',')
                combine += attr
                for index, string in enumerate(combine):
                    combine[index] = string.lstrip().rstrip()
                for i, s in enumerate(combine):
                    if not s or s in IGNORE_BLANKS:
                        del combine[i]
        return combine

    @staticmethod
    def __convert_num(number):
        """
        All path items are automatically strings. If you think it's an int or float, this attempts to convert it.
        :param number (str):
        :return: (float)
        """
        try:
            return float(number)
        except ValueError:
            return number

    @staticmethod
    def __camel_case(word):
        """
        Convert underscore naming into camel case naming
        :param word (str):
        :return: (str)
        """
        word = word.lower()
        if '_' in word:
            split_word = word.split('_')
        else:
            split_word = word.split()
        if len(split_word) > 0:
            for i, word in enumerate(split_word):
                if i > 0:
                    split_word[i] = word.title()
        strings = ''.join(split_word)
        return strings

    @staticmethod
    def __name_unit_regex(word):
        """
        Split a name and unit that are bunched together (i.e. '250m')
        :param word (str):
        :return:
        """
        r = re.findall(r'(\d+)(\w+?)', word)
        value = r[0][0]
        unit = r[0][1]
        try:
            value = float(value)
        except ValueError:
            pass
        return value, unit

    @staticmethod
    def __contains_digits(word):
        """
        Check if the string contains digits
        :param word (str):
        :return:
        """
        return any(i.isdigit() for i in word)

    def __split_name_unit(self, line):
        """
        Split a string that has value and unit as one.
        :param line (str):
        :return:
        """
        if line != '' or line != ' ':
            # If there are parenthesis, remove them
            line = line.replace('(', '').replace(')', '')
            # When value and units are a range (i.e. '100 m - 200 m').
            if ' to ' in line or ' - ' in line:
                line = line.replace('to', '').replace('-', '')
                val_list = [int(s) for s in line.split() if s.isdigit()]
                unit_list = [s for s in line.split() if not s.isdigit()]
                # For items that did not split properly. Need regex split.
                for item in unit_list:
                    if self.__contains_digits(item):
                        unit_list = []
                        i, v = self.__name_unit_regex(item)
                        val_list.append(i)
                        unit_list.append(v)
                # Piece the number range back together.
                value = str(val_list[0]) + ' to ' + str(val_list[1])
                unit = unit_list[0]
            else:
                # Normal case. Value and unit separated by a space.
                if ' ' in line:
                    line = line.split()
                    value = line[0]
                    unit = ' '.join(line[1:])
                # No Value. Line only contains a unit.
                elif not self.__contains_digits(line):
                    value = 'n/a'
                    unit = line
                # Value and unit bunched together ('100m'). Use regex to identify groups.
                else:
                    value, unit = self.__name_unit_regex(line)
            return value, unit

    @staticmethod
    def __str_cleanup(line):
        """
        Remove the unnecessary characters in the line that we don't want
        :param line (str):
        :return: (str)
        """
        line = line.rstrip()
        if '#' in line:
            line = line.replace("#", "")
            line = line.lstrip()
        if '-----------' in line:
            line = ''

        return line

    @staticmethod
    def __slice_key_val(line):
        """
        Get the key and value items from a line by looking for and lines that have a ":"
        :param line (str):
        :return: (str) Key, (str) Value
        """
        position = line.find(":")
        # If value is -1, that means the item was not found in the string.
        if position != -1:
            key = line[:position]
            value = line[position+1:]
            value = value.lstrip()
            return key, value
        else:
            key = line
            value = None
            return key, value

    def __create_coordinates(self, lat, lon):
        """
        GeoJSON standard:
        Use to determine 2-point or 4-point coordinates
        :param lat (float):
        :param lon (float):
        :return: (dict)
        """

        # Sort lat an lon in numerical order
        lat.sort()
        lon.sort()
        # 4 coordinate values
        if len(lat) == 2 and len(lon) == 2:
            geo_dict = self.__geo_multipoint(lat, lon)
        # 2 coordinate values
        elif len(lat) == 1 and len(lon) == 1:
            geo_dict = self.__geo_point(lat, lon)
        else:
            geo_dict = {}
            print("More than 4 coordinates")
        return geo_dict

    def __geo_multipoint(self, lat, lon):
        """
        GeoJSON standard:
        Create a geoJson MultiPoint-type dictionary
        :param lat (float):
        :param lon (float):
        :return:
        """
        geo_dict = OrderedDict()
        geometry_dict = OrderedDict()
        coordinates = []
        # bbox = []
        temp = [None, None]

        # if the value pairs are matching, then it's not a real MultiPoint type. Send to other method
        if lat[0] == lat[1] and lon[0] == lon[1]:
            lat.pop()
            lon.pop()
            geo_dict = self.__geo_point(lat, lon)

        # 4 unique values
        else:
            # # Creates bounding box
            # for index, point in enumerate(lat):
            #     bbox.append(lat[index])
            #     bbox.append(lon[index])

            # Creates coordinates list
            for i in lat:
                temp[0] = i
                for j in lon:
                    temp[1] = j
                    coordinates.append(copy.copy(temp))

            # Create geometry block
            geometry_dict['type'] = 'Polygon'
            geometry_dict['coordinates'] = coordinates

            # Create geo block
            geo_dict['type'] = 'Featured'
            # geo_dict['bbox'] = bbox
            geo_dict['geometry'] = geometry_dict

        return geo_dict

    @staticmethod
    def __geo_point(lat, lon):
        """
        GeoJSON standard:
        Create a geoJson Point-type dictionary
        :param lat:
        :param lon:
        :return:
        """
        coordinates = []
        geo_dict = OrderedDict()
        geometry_dict = OrderedDict()
        for index, point in enumerate(lat):
            coordinates.append(lat[index])
            coordinates.append(lon[index])
        geometry_dict['type'] = 'Point'
        geometry_dict['coordinates'] = coordinates
        geo_dict['type'] = 'Feature'
        geo_dict['geometry'] = geometry_dict
        return geo_dict





"""
GLOBAL LIST OF EMPTIES AND IGNORED TERMS
"""

# GENERAL
GEN_EMPTY = ['', ' ', None, 'na', 'n/a', 'nan', '?', "'", "''"]

# EXCEL
EXCEL_EMPTY = ['', ' ', None, 'na', 'nan']

# TimeSeries Convert
EMPTY = ['', ' ', None, 'na', 'n/a', 'nan', '?']

# NOAA
NOAA_EMPTY = ['\n', '', '#\n', '# \n', ' ']
NOAA_KEYS = ['earliest_year', 'most_recent_year', 'note']
NOAA_VAR_LINES = ['have no #', 'variables', 'lines begin with #', 'double marker', 'line variables format',
                    'line format', 'c or n for character or numeric', 'preceded by']
NOAA_DATA_LINES = ['have no #', 'tab-delimited text', 'age ensembles archived']


# DOI





import random 
import datetime

def what_is_my_name():
    """
    I do not know your name, but this function will be helpful for learning!
    :return: None
    """
    print("I have no idea.")
    return

def what_time_is_it():
    """
    Prints out the current time and day.
    :return: None
    """
    print(str(datetime.datetime.now().strftime("%d %B %Y %I:%M%p")))
    return

def tell_me_a_joke():
    """
    Randomly selects a joke and prints it to the console. 
    :returns: None
    """
    jokes = [
        "A man didn't like his haircut, but it started to grow on him.",
        "Q: What did one snowman say to the other? A: Do you smell carrots?",
        "There’s two fish in a tank. One turns to the other and says 'You man the guns, I’ll drive’",
        "Some jokes just fluorite over my head",
        "Q: Why do Earth Science professors like to teach about ammonia? A: Because it's basic material."
    ]
    
    print(jokes[random.randint(0, len(jokes))])
    return


import csv


def read_csv_to_columns(filename):
    """
    Opens the target CSV file and creates a dictionary with one list for each CSV column.
    :param filename: (str) Filename
    :return: (dict) CSV data. Keys: Column number(int), Values: Column data (list)
    """
    d = {}
    try:
        with open(filename, 'r') as f:
            r = csv.reader(f, delimiter=',')
            # Create a dict with X lists corresponding to X columns
            for idx, col in enumerate(next(r)):
                d[idx] = []
            # Start iter through CSV data
            for row in r:
                for idx, col in enumerate(row):
                    # Append the cell to the correct column list
                    try:
                        d[idx].append(float(col))
                    except ValueError:
                        d[idx].append(col)
    except FileNotFoundError:
        print('CSV FileNotFound: ' + filename)
    return d


def write_csv_to_file(filename, d):
    """
    Writes columns of data to a target CSV file.
    :param filename: (str) Target CSV file
    :param d: (dict) A dictionary containing one list for every data column. Keys: int, Values: list
    :return: None
    """
    l_columns = []
    for k, v in d.items():
        l_columns.append(v)
    rows = zip(*l_columns)
    with open(filename, 'w+') as f:
        w = csv.writer(f)
        for row in rows:
            w.writerow(row)
    return


def add_csv_to_json(d):
    """
    Using the given paleoData dictionary from the JSON metadata, retrieve CSV data from CSV files, and insert the CSV
    data columns to their respective JSON paleoData table columns.
    :param d: (dict) PaleoData dictionary
    :return: (dict) Modified original dictionary (dict) CSV column data
    """
    d2 = {}
    # Loop through each table in paleoData
    for table_name, table_content in d.items():
        # Create CSV entry into dictionary that contains all columns.
        d2[table_content['filename']] = read_csv_to_columns(table_content['filename'])
        # Start putting CSV data into corresponding JSON metadata columns under 'values' key.
        for col_name, col_content in table_content['columns'].items():
            col_content['values'] = d2[table_content['filename']][col_content['number']-1]
    return d, d2



import re

"""
GLOBAL LIST OF REGEXES
"""

# DOI
DOI = re.compile(r'\b(10[.][0-9]{3,}(?:[.][0-9]+)*/(?:(?!["&\'<>,])\S)+)\b')

# NOAA
# Convert camelCase to underscore
first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')

# TimeSeries Convert
re_misc_fetch = re.compile(r'(geo_(\w+)|climateInterpretation_(\w+)|calibration_(\w+)|paleoData_(\w+))')
re_pub_fetch = re.compile(r'pub1_(citation|year|DOI|author|publisher|title|type|volume|issue|journal|link|pubDataUrl|abstract|pages)')

re_pub_valid = re.compile(r'pub(\d)_(citation|year|DOI|author|publisher|title|type|volume|issue|journal|link|pubDataUrl|abstract|pages)')
re_fund_valid = re.compile(r'funding(\d)_(grant|agency)')

re_pub_invalid = re.compile(r'pub_(\w+)|pub(\d)_(\w+)|pub(\d)(\w+)|pub(\w+)')
re_fund_invalid = re.compile(r'agency|grant|funding_agency|funding_grant')
re_geo_invalid = re.compile(r'geo(\w+)|geo_(\w+)')
re_paleo_invalid = re.compile(r'paleodata(\w+)|paleodata_(\w+)|measurement(\w+)|measurement_(\w+)')
re_calib_invalid = re.compile(r'calibration(\w+)|calibration_(\w+)')
re_clim_invalid = re.compile(r'climateinterpretation(\w+)|climateinterpretation_(\w+)')

re_pub_nh = re.compile(r'pub(\d)_(\w+)')
re_pub_cc = re.compile(r'pub(\w+)')
re_pub_h = re.compile(r'pub_(\w+)')
re_pub_n = re.compile(r'pub(\d)(\w+)')


import os

from ..helpers.directory import list_files, create_tmp_dir
from ..helpers.jsons import read_json_from_file
from ..helpers.zips import *
from .lpd_noaa import LPD_NOAA
from .noaa_lpd import NOAA_LPD


def noaa():
    """
    Convert between NOAA and LiPD file formats.
    :return:
    """
    # Run lpd_noaa or noaa_lpd ?
    print("Which conversion?\n1. LPD to NOAA\n2. NOAA to LPD\n")
    mode = input("Option: ")
    f_list = []
    ft = ''

    # .lpd to noaa
    if mode == '1':
        f_list = list_files('.lpd')
        ft = ' LiPD'
    # Find all needed files in current directory
    elif mode == '2':
        f_list = list_files('.txt')
        if 'noaa-blank.txt' in f_list:
            f_list.remove('noaa-blank.txt')
        if 'quarantine.txt' in f_list:
            f_list.remove('quarantine.txt')
        ft = ' NOAA'

    print("Found {0} {1} file(s)".format(str(len(f_list)), ft))
    # Process each available file
    for name_ext in f_list:
        print('processing: {}'.format(name_ext))

        # File name w/o extension
        name = os.path.splitext(name_ext)[0]

        # Unzip file and get tmp directory path
        dir_tmp = create_tmp_dir()

        # Process file
        if mode == '1':
            unzip(name_ext, dir_tmp)
            _process_lpd(name, dir_tmp)

        elif mode == '2':
            _process_noaa(name, dir_tmp)

        # Delete tmp folder and all contents
        shutil.rmtree(dir_tmp)

    print("Process Complete")

    return


def _process_noaa(name, dir_tmp):
    """
    Convert NOAA format to LiPD format
    :param name: (str) Name of file, no extension
    :param dir_tmp: (str) Path to tmp directory
    :return: none
    """
    # dir_root = os.getcwd()
    dir_root = os.path.expanduser('~/LiPD_Notebooks')
    NOAA_LPD(dir_root, dir_tmp, name).main()

    return


def _process_lpd(name, dir_tmp):
    """
    Convert a LiPD format to NOAA format
    :param name: (str) Name of file, no extension
    :param dir_tmp: (str) Path to tmp directory
    :return: none
    """

    dir_root = os.getcwd()
    dir_bag = os.path.join(dir_tmp, name)
    dir_data = os.path.join(dir_bag, 'data')

    # Navigate down to jld file
    # Directory Change: dir_root -> dir_data
    os.chdir(dir_data)

    # Open file and execute conversion script
    d = read_json_from_file(os.path.join(dir_data, name + '.jsonld'))
    LPD_NOAA(dir_root, name, d).main()

    # except ValueError:
    #     txt_log(dir_root, 'quarantine.txt', name, "Invalid Unicode characters. Unable to load file.")

    # Move back to root for next loop
    # Directory Change: dir_data -> dir_root
    os.chdir(dir_root)

    return




import re
import urllib.error
from collections import OrderedDict

import requests

from ..helpers.loggers import *
from ..helpers.jsons import *


class DOIResolver(object):
    """
    Use DOI id(s) to pull updated publication info from doi.org and overwrite file data.

    Input:  Original publication dictionary
    Output: Updated publication dictionary (success), original publication dictionary (fail)
    """

    def __init__(self, dir_root, name, root_dict):
        """
        :param dir_root: (str) Path to dir containing all .lpd files
        :param name: (str) Name of current .lpd file
        :param root_dict: (dict) Full dict loaded from jsonld file
        """
        self.dir_root = dir_root
        self.name = name
        self.root_dict = root_dict

    def main(self):
        """
        Main function that gets file(s), creates outputs, and runs all operations.
        :return: (dict) Updated or original data for jsonld file
        """

        for idx, pub in enumerate(self.root_dict['pub']):
            # Retrieve DOI id key-value from the root_dict
            doi_string, doi_found = self.find_doi(pub)

            if doi_found:
                # Empty list for no match, or list of 1+ matching DOI id strings
                doi_list = self.clean(doi_string)
                if not doi_list:
                    self.illegal_doi(doi_string)
                else:
                    for doi_id in doi_list:
                        self.get_data(doi_id, idx)
            else:
                # Quarantine the flagged file and log it
                txt_log(self.dir_root, self.name, "quarantine.txt", "Publication #" + str(idx) + ": DOI not provided")
                self.root_dict['pub'][idx]['pubDataUrl'] = 'Manually Entered'
            # self.remove_empties(idx)
        return remove_empty_fields(self.root_dict)

    @staticmethod
    def clean(doi_string):
        """
        Use regex to extract all DOI ids from string (i.e. 10.1029/2005pa001215)
        :param doi_string: (str) Raw DOI string value from input file. Often not properly formatted.
        :return: (list) DOI ids. May contain 0, 1, or multiple ids.
        """
        regex = re.compile(r'\b(10[.][0-9]{3,}(?:[.][0-9]+)*/(?:(?!["&\'<>,])\S)+)\b')
        try:
            # Returns a list of matching strings
            m = re.findall(regex, doi_string)
        except TypeError:
            # If doi_string is None type, return empty list
            m = []
        return m

    @staticmethod
    def compile_date(date_parts):
        """
        Compiles date only using the year
        :param date_parts: List of date parts retrieved from doi.org
        :return: Date string or NaN
        """
        if date_parts[0][0]:
            return date_parts[0][0]
        return 'NaN'

    @staticmethod
    def compile_authors(authors):
        """
        Compiles authors "Last, First" into a single list
        :param authors: (list) Raw author data retrieved from doi.org
        :return: (list) Author objects
        """
        author_list = []
        for person in authors:
            author_list.append({'name': person['family'] + ", " + person['given']})
        return author_list

    @staticmethod
    def compare_replace(pub_dict, fetch_dict):
        """
        Take in our Original Pub, and Fetched Pub. For each Fetched entry that has data, overwrite the Original entry
        :param pub_dict: (dict) Original pub dictionary
        :param fetch_dict: (dict) Fetched pub dictionary from doi.org
        :return: (dict) Updated pub dictionary, with fetched data taking precedence
        """
        blank = [" ", "", None]
        for k, v in fetch_dict.items():
            try:
                if fetch_dict[k] != blank:
                    pub_dict[k] = fetch_dict[k]
            except KeyError:
                pass
        return pub_dict

    def remove_empties(self, pub):
        for x in list(self.root_dict['pub'][pub].keys()):
            if x == 'identifier':
                if self.root_dict['pub'][pub][x][0]['id'] in EMPTY:
                    del self.root_dict['pub'][pub][x]
            elif self.root_dict['pub'][pub][x] in EMPTY:
                del self.root_dict['pub'][pub][x]
        return

    def noaa_citation(self, doi_string):
        """
        Special instructions for moving noaa data to the correct fields
        :param doi_string: (str) NOAA url
        :return: None
        """
        # Append location 1
        if 'link' in self.root_dict['pub'][0]:
            self.root_dict['pub'][0]['link'].append({"url": doi_string})
        else:
            self.root_dict['pub'][0]['link'] = [{"url": doi_string}]

        # Append location 2
        self.root_dict['dataURL'] = doi_string

        return

    def illegal_doi(self, doi_string):
        """
        DOI string did not match the regex. Determine what the data is.
        :param doi_string: (str) Malformed DOI string
        :return: None
        """

        # Ignores empty or irrelevant strings (blank, spaces, na, nan, ', others)
        if len(doi_string) > 5:

            # NOAA string
            if 'noaa' in doi_string.lower():
                self.noaa_citation(doi_string)

            # Paragraph citation / Manual citation
            elif doi_string.count(' ') > 3:
                self.root_dict['pub'][0]['citation'] = doi_string

            # Strange Links or Other, send to quarantine
            else:
                txt_log(self.dir_root, self.name, "quarantine.txt", "Malformed DOI: " + doi_string)

        return

    def compile_fetch(self, raw, doi_id):
        """
        Loop over Raw and add selected items to Fetch with proper formatting
        :param raw: (dict) JSON data from doi.org
        :param doi_id: (str)
        :return:
        """
        fetch_dict = OrderedDict()
        order = {'author': 'author', 'type': 'type', 'identifier': '', 'title': 'title', 'journal': 'container-title',
                'pubYear': '', 'volume': 'volume', 'publisher': 'publisher', 'page':'page', 'issue': 'issue'}

        for k, v in order.items():
            try:
                if k == 'identifier':
                    fetch_dict[k] = [{"type": "doi", "id": doi_id, "url": "http://dx.doi.org/" + doi_id}]
                elif k == 'author':
                    fetch_dict[k] = self.compile_authors(raw[v])
                elif k == 'pubYear':
                    fetch_dict[k] = self.compile_date(raw['issued']['date-parts'])
                else:
                    fetch_dict[k] = raw[v]
            except KeyError as e:
                # If we try to add a key that doesn't exist in the raw dict, then just keep going.
                pass

        return fetch_dict

    def get_data(self, doi_id, idx):
        """
        Resolve DOI and compile all attributes into one dictionary
        :param doi_id: (str)
        :param idx: (int) Publication index
        :return: (dict) Updated publication dictionary
        """

        tmp_dict = self.root_dict['pub'][0].copy()
        try:
            # Send request to grab metadata at URL
            url = "http://dx.doi.org/" + doi_id
            headers = {"accept": "application/rdf+xml;q=0.5, application/citeproc+json;q=1.0"}
            r = requests.get(url, headers=headers)

            # DOI 404. Data not retrieved. Log and return original pub
            if r.status_code == 400:
                txt_log(self.dir_root, self.name, "quarantine.txt", "DOI.org 404 response: " + doi_id)

            # Ignore other status codes. Run when status is 200 (good response)
            elif r.status_code == 200:

                # Load data from http response
                raw = json.loads(r.text)

                # Create a new pub dictionary with metadata received
                fetch_dict = self.compile_fetch(raw, doi_id)

                # Compare the two pubs. Overwrite old data with new data where applicable
                tmp_dict = self.compare_replace(tmp_dict, fetch_dict)
                tmp_dict['pubDataUrl'] = 'doi.org'
                self.root_dict['pub'][idx] = tmp_dict

        except urllib.error.URLError:
            txt_log(self.dir_root, self.name, "quarantine.txt", "Malformed DOI: " + doi_id)
        except ValueError:
            txt_log(self.dir_root, self.name, "quarantine.txt", "Cannot resolve DOIs from this publisher: " + doi_id)

        return

    def find_doi(self, curr_dict):
        """
        Recursively search the file for the DOI id. More taxing, but more flexible when dictionary structuring isn't absolute
        :param curr_dict: (dict) Current dictionary being searched
        :recursive return: (dict) Current dictionary, (bool) False flag that DOI was not found
        :final return: (str) DOI id, (bool) True flag that DOI was found
        """
        try:
            if 'id' in curr_dict:
                return curr_dict['id'], True
            elif isinstance(curr_dict, list):
                for i in curr_dict:
                    return self.find_doi(i)
            elif isinstance(curr_dict, dict):
                for k, v in curr_dict.items():
                    if k == 'identifier':
                        return self.find_doi(v)
                return curr_dict, False
            else:
                return curr_dict, False
        # If the ID key doesn't exist, then return the original dict with a flag
        except TypeError:
            return curr_dict, False



import os
import tempfile
import shutil
import ntpath
import time
import tkinter
from tkinter import filedialog


def file_from_path(path):
    """
    Extract the file name from a given file path.
    :param path: (str) File path
    :return: (str) File name with extension
    """
    head, tail = ntpath.split(path)
    return head, tail or ntpath.basename(head)


def create_tmp_dir():
    """
    Creates tmp directory in OS temp space.
    :return: (str) Path to tmp directory
    """
    return tempfile.mkdtemp()


def list_files(x):
    """
    Lists file(s) in given path of the X type.
    :param x: (str) File extension that we are interested in.
    :return: (list of str) File name(s) to be worked on
    """
    file_list = []
    for file in os.listdir():
        if file.endswith(x):
            file_list.append(file)
    return file_list


def dir_cleanup(dir_bag, dir_data):
    """
    Moves JSON and csv files to bag root, then deletes all the metadata bag files. We'll be creating a new bag with
    the data files, so we don't need the other text files and such.
    :param dir_bag: (str) Path to root of Bag
    :param dir_data: (str) Path to Bag /data subdirectory
    :return: None
    """
    # dir : dir_data -> dir_bag
    os.chdir(dir_bag)

    # Delete files in dir_bag
    for file in os.listdir(dir_bag):
        if file.endswith('.txt'):
            os.remove(os.path.join(dir_bag, file))

    # Move dir_data files up to dir_bag
    for file in os.listdir(dir_data):
        shutil.move(os.path.join(dir_data, file), dir_bag)

    # Delete empty dir_data folder
    shutil.rmtree(dir_data)

    return


def check_file_age(filename, days):
    """
    Check if the target file has an older creation date than X amount of time.
    i.e. One day: 60*60*24
    :param filename: (str) Target filename
    :param days: (int) Limit in number of days
    :return: (bool) True - older than X time, False - not older than X time
    """
    # Multiply days given by time for one day.
    t = days * 60 * 60 * 24
    now = time.time()
    specified_time = now - t
    try:
        if os.path.getctime(filename) < specified_time:
            # File found and out of date
            return True
        # File found, and not out of date
        return False
    except FileNotFoundError:
        # File not found. Need to download it.
        return True


def browse_dialog():
    """
    Open up a GUI browse dialog window and let to user pick a target directory.
    :return str: Target directory path
    """

    root = tkinter.Tk()
    root.withdraw()
    root.update()
    path = tkinter.filedialog.askdirectory(parent=root, initialdir=os.path.expanduser('~'), title='Please select a directory')
    root.destroy()

    return path


def set_source():
    """
    User sets the path to LiPD source. Local or online.
    :return: (str) Path
    """
    path = None
    invalid = True
    count = 0
    print("Where are your files stored? Choose an option by number:\n1. Online URL\n2. Local Computer\n3. "
      "Downloads "
      "Folder\n")
    while invalid:
        option = input("Option: ")
        if option == '1':
            # Retrieve data from the online URL
            path = input("Enter the URL: ")
        elif option == '2':
            # Open up the GUI browse dialog
            path = browse_dialog()
            # Set the path to the local files in CLI and lipd_lib
        elif option == '3':
            # Set the path to the system downloads folder.
            path = os.path.expanduser('~/Downloads')
        else:
            # Something went wrong. Prompt again. Give a couple tries before defaulting to downloads folder
            if count == 2:
                print("Defaulting to Downloads Folder.")
                path = os.path.expanduser('~/Downloads')
            else:
                count += 1
                print("Invalid option. Try again.")
        if path:
            invalid = False
    print(path)
    return path



import copy
import json
import os

import demjson

EMPTY = ['', ' ', None, 'na', 'n/a', 'nan', '?', "'", "''"]


def write_json_to_file(filename, json_data):
    """
    Write all JSON in python dictionary to a new json file.
    :param filename: (str) Target File (name + .json)
    :param json_data: (dict) JSON data
    :return: None
    """
    # Use demjson to maintain unicode characters in output
    json_bin = demjson.encode(json_data, encoding='utf-8', compactly=False)
    # Write json to file
    try:
        open(filename, "wb").write(json_bin)
    except FileNotFoundError:
        print("Error writing json: " + filename)
    return


def read_json_from_file(filename):
    """
    Import the JSON data from target file.
    :param filename: (str) Target Files
    :return: (dict) JSON data
    """
    d = {}
    try:
        # Open json file and read in the contents. Execute DOI Resolver?
        # with open(filename, 'r') as f:
            # Load json into dictionary
        d = demjson.decode_file(filename)
    except FileNotFoundError:
        try:
            d = demjson.decode_file(os.path.splitext(filename)[0] + '.json')
        except FileNotFoundError:
            print("Read Json: File not found")
    return d


def remove_csv_from_json(d):
    """
    Remove all CSV data 'values' entries from paleoData table in the JSON structure.
    :param d: (dict) JSON data - old structure
    :return: (dict) Metadata dictionary without CSV values
    """
    # Loop through each table in paleoData
    for table, table_content in d['paleoData'].items():
        for column, column_content in table_content['columns'].items():
            try:
                # try to delete the values key entry
                del column_content['values']
            except KeyError:
                # if the key doesn't exist, keep going
                print("Remove CSV from JSON: Error removing column values")
    return d


def get_csv_from_json(d):
    """
    Get CSV values when mixed into json data. Pull out the CSV data and put it into a dictionary.
    :param d: (dict) JSON with CSV values
    :return: (dict) CSV values. (i.e. { CSVFilename1: { Column1: [Values], Column2: [Values] }, CSVFilename2: ... }
    """
    csv = {}
    try:
        for table, table_content in d['paleoData'].items():
            # Create entry for this table/CSV file (i.e. Asia-1.measTable.PaleoData.csv)
            # Note: Each table has a respective CSV file.
            csv[table_content['filename']] = {}
            for column, column_content in table_content['columns'].items():
                # Set the "values" into csv dictionary in order of column "number"
                csv[table_content['filename']][column_content['number']] = column_content['values']
    except KeyError:
        print("Get CSV from JSON: KeyError")

    return csv


def remove_empty_fields(d):
    """
    Go through N number of nested data types and remove all empty entries. Recursion
    :param d: (any) Dictionary, List, or String of data
    :return: (any) Returns a same data type as original, but without empties.
    """
    # Int types don't matter. Return as-is.
    if not isinstance(d, int):
        if isinstance(d, str) or d is None:
            try:
                # Remove new line characters and carriage returns
                d = d.rstrip()
            except AttributeError:
                # None types don't matter. Keep going.
                pass
            if d in EMPTY:
                # Substitute empty entries with ""
                d = ''
        elif isinstance(d, list):
            # Recurse once for each item in the list
            for i, v in enumerate(d):
                d[i] = remove_empty_fields(d[i])
            # After substitutions, remove and empty entries.
            for i in d:
                if not i:
                    d.remove(i)
        elif isinstance(d, dict):
            # First, go through and substitute "" (empty string) entry for any values in EMPTY
            for k, v in d.items():
                d[k] = remove_empty_fields(v)
            # After substitutions, go through and delete the key-value pair.
            # This has to be done after we come back up from recursion because we cannot pass keys down.
            for key in list(d.keys()):
                if not d[key]:
                    del d[key]

    return d


def remove_empty_doi(d):
    """
    If an "identifier" dictionary has no doi ID, then it has no use. Delete it.
    :param d: (dict) JSON Metadata
    :return: (dict) JSON Metadata
    """
    try:
        # Check each publication dictionary
        for pub in d['pub']:
            # If no identifier, then we can quit here. If identifier, then keep going.
            if 'identifier' in pub:
                if 'id' in pub['identifier'][0]:
                    # If there's a DOI id, but it's EMPTY
                    if pub['identifier'][0]['id'] in EMPTY:
                        del pub['identifier']
                else:
                    # If there's an identifier section, with no DOI id
                    del pub['identifier']
    except KeyError:
        # What else could go wrong?
        pass

    return d


def old_to_new_structure(d):
    """
    Restructure JSON to the new format. Table and columns referenced by name, not index. (dictionaries)
    :param d:(dict) JSON metadata
    :return: (dict) JSON metadata
    """
    # PaleoData dict
    tmp_p = {}

    for table in d['paleoData']:
        # All tables, all columns, table name
        tmp_t = {}
        tmp_c = {}
        t_name = table['paleoDataTableName']

        # paleoDataTableName and variableName are current
        # tableName and parameter are deprecated

        # Restructure columns into tmp_c
        for col in table['columns']:
            c_name = col['variableName']
            tmp_c[c_name] = col

        # Move table strings into tmp_t
        for k, v in table.items():
            if isinstance(v, str):
                tmp_t[k] = v

        # Move new column structure to table
        tmp_t['columns'] = copy.deepcopy(tmp_c)

        # Move table into tmp_p
        tmp_p[t_name] = copy.deepcopy(tmp_t)

    # Overwrite original paleoData dictionary with new dictionary
    d['paleoData'] = tmp_p
    return d


def new_to_old_structure(d):
    """
    Restructure JSON to the old format. Table and columns are indexed by numbers. (lists)
    :param d:(dict) JSON metadata
    :return: (dict) JSON metadata
    """
    # PaleoData dict
    tmp_p = []

    for k, v in d['paleoData'].items():
        tmp_c = []
        # For each column, append the value, forget the key.
        for i, e in v['columns'].items():
            tmp_c.append(e)
        # Replace the old column dict with the new column list
        v['columns'] = copy.deepcopy(tmp_c)
        # Append the table to the PaleoData Dict to the list, forget the key.
        tmp_p.append(v)

    # Overwrite original paleoData dictionary with new dictionary
    d['paleoData'] = tmp_p
    return d


def split_csv_json(d):
    """
    Split JSON with CSV values into separate JSON and CSV dictionaries.
    :param d: (dict) JSON metadata with CSV values in paleoData columns
    :return: (dict) JSON only metadata (dict) CSV organized by filename->column
    """
    # First, get CSV values and organize.
    csv = get_csv_from_json(d)
    # Then remove CSV values, which gives us JSON only.
    jsn = remove_csv_from_json(d)
    return jsn, csv






import bagit


def create_bag(dir_bag):
    """
    Create a Bag out of given files.
    :param dir_bag: (str) Directory that contains csv, jsonld, and changelog files.
    :return: (obj) Bag
    """
    bag = bagit.make_bag(dir_bag, {'Name': 'LiPD Project', 'Reference': 'www.lipd.net', 'DOI-Resolved': 'True'})
    return bag


def open_bag(dir_bag):
    """
    Open Bag at the given path
    :param dir_bag: (str) Path to Bag
    :return: (obj) Bag
    """
    bag = bagit.Bag(dir_bag)
    return bag


def validate_md5(bag):
    """
    Check if Bag is valid
    :param bag: (obj) Bag
    :return: None
    """
    if bag.is_valid():
        print("Valid md5")
        # for path, fixity in bag.entries.items():
        #     print("path:{}\nmd5:{}\n".format(path, fixity["md5"]))
    else:
        print("Invalid md5")
    return


def resolved_flag(bag):
    """
    Check DOI flag in bag.info to see if doi_resolver has been previously run
    :param bag: (obj) Bag
    :return: (bool) Flag
    """
    if 'DOI-Resolved' in bag.info:
        return True
    return False




from setuptools import setup, find_packages
from os import path
import shutil
import os
import io
import sys
from distutils.command.install import install


class PostInstall(install):
    """ Custom install script that runs post-install."""
    def run(self):
        # Make the notebooks folder in the user directory
        src_nb = os.path.join(here, 'lipd', 'files')
        src_ex = os.path.join(src_nb, 'examples')
        dst_nb = os.path.join(os.path.expanduser('~'), 'LiPD_Notebooks')
        dst_ex = os.path.join(dst_nb, 'examples')

        # Make folders if needed
        if not os.path.isdir(dst_nb):
            os.mkdir(dst_nb)
        if not os.path.isdir(dst_ex):
            os.mkdir(dst_ex)

        # Copy example files (don't overwrite directory)
        f = os.listdir(src_ex)
        for file in f:
            if os.path.isfile(file):
                shutil.copy(file, src_ex)

        # Copy / Overwrite Quickstart notebook
        shutil.copy(os.path.join(src_nb, 'Welcome LiPD - Quickstart.ipynb'), dst_nb)

        # Open the install folder so the user can see the documentation and instructions
        os.system('open .')
        install.run(self)


here = path.abspath(path.dirname(__file__))
version = '0.1.1.9'

# Read the readme file contents into variable
if sys.argv[-1] == 'publish' or sys.argv[-1] == 'publishtest':
    os.system('pandoc README.md -f markdown -t rst -s -o README.txt')

readme_file = io.open('README.txt', encoding='utf-8')
# Fallback long_description in case errors with readme file.
long_description = "Welcome to LiPD. Please reference the README file in the package for information"
with readme_file:
    long_description = readme_file.read()

# Publish the package to the live server
if sys.argv[-1] == 'publish':
    # Register the tarball, upload it, and trash the temp readme rst file
    os.system('python3 setup.py register')
    os.system('python3 setup.py sdist upload')
    os.remove('README.txt')
    sys.exit()

# Publish the package to the test server
elif sys.argv[-1] == 'publishtest':
    # Create dist tarball, register it to test site, upload tarball, and remove temp readme file
    os.system('python3 setup.py sdist')
    os.system('python3 setup.py register -r https://testpypi.python.org/pypi')
    os.system('twine upload -r test dist/LiPD-' + version + '.tar.gz')
    # Trash the temp rst readme file
    os.remove('README.txt')
    sys.exit()


# Do all the setup work
setup(
    name='LiPD',
    version=version,
    author='C. Heiser',
    author_email='heiser@nau.edu',
    packages=find_packages(exclude=['build', '_docs', 'templates']),
    # packages = ["lipd", "doi", "noaa", "excel"],
    entry_points={
        "console_scripts": [
            'lipd= lipd.__main__:main'
        ]
    },
    cmdclass={
        'install': PostInstall,
    },
    url='https://github.com/nickmckay/LiPD-utilities',
    license='GNU Public',
    description='LiPD utilities to process, convert, and analyze data.',
    long_description=long_description,
    keywords="paleo R matlab python paleoclimatology linkedearth",
    install_requires=[
        "bagit>=1.5.4",
        "beautifulsoup4>=4.4.1",
        "demjson>=2.2.4",
        "matplotlib>=1.4.2",
        "xlrd>=0.9.3",
        "Pillow>=3.1.1",
        "jupyter>=1.0.0",
        "requests>=2.9.1",
        "google-api-python-client>=1.4.2",
        "virtualenv"
    ],
    include_package_data=False,
    package_data={},
)




import os
import datetime


def update_changelog():
    """
    Create or update the changelog txt file. Prompt for update description.
    :return: none
    """
    # description = input("Please enter a short description for this update:\n ")
    description = 'Placeholder for description here.'

    # open changelog file for appending. if doesn't exist, creates file.
    with open('changelog.txt', 'a+') as f:
        # write update line
        f.write(str(datetime.datetime.now().strftime("%d %B %Y %I:%M%p")) + '\nDescription: ' + description)
    return


def txt_log(dir_root, filename, quarantine_txt, info):
    """
    Debug Log. Log names and error of problematic files to a txt file
    :param dir_root: (str) Directory containing .lpd file(s)
    :param quarantine_txt: (str) Name of the txt file to be written to
    :param filename: (str) Name of the current .lpd file
    :param info: (str) Error description
    :return: None
    """
    org_dir = os.getcwd()
    os.chdir(dir_root)
    with open(quarantine_txt, 'a+') as f:
        try:
            # Write update line
            f.write("File: " + filename + "\n" + "Error: " + info + "\n\n")
        except KeyError:
            print("Debug Log Error")
    os.chdir(org_dir)
    return


def txt_log_end(dir_root, quarantine_txt):
    """
    At the end of a batch, add a divider line and timestamp
    Quarantine.txt file is a continually growing file.
    :param dir_root: (str) Directory containing .lpd file(s)
    :param quarantine_txt: (str) Name of the txt file to be written to
    :return: None
    """
    org_dir = os.getcwd()
    os.chdir(dir_root)
    with open(quarantine_txt, 'a+') as f:
        try:
            # Write update line
            f.write('Timestamp: {:%Y-%m-%d %H:%M:%S}'.format(datetime.datetime.now()))
            f.write("\n------------------------------------\n")
        except KeyError:
            print("Debug Log Error")
    os.chdir(org_dir)
    return



from .doi_resolver import DOIResolver
from ..helpers.bag import *
from ..helpers.directory import *
from ..helpers.zips import unzip, re_zip
from ..helpers.jsons import json, write_json_to_file
from ..helpers.loggers import txt_log, update_changelog


def doi():
    """
    Main function that controls the script. Take in directory containing the .lpd file(s). Loop for each file.
    :param dir_root: (str) Directory location of target files
    :return: None
    """

    # Find all .lpd files in current directory
    # dir: ? -> dir_root
    dir_root = os.getcwd()
    f_list = list_files('.lpd')
    print("Found {0} {1} file(s)".format(str(len(f_list)), 'LiPD'))

    for name_ext in f_list:
        print('processing: {}'.format(name_ext))

        # .lpd name w/o extension
        name = os.path.splitext(name_ext)[0]

        # Unzip file and get tmp directory path
        dir_tmp = create_tmp_dir()
        unzip(name_ext, dir_tmp)

        # Unbag and check resolved flag. Don't run if flag exists
        if resolved_flag(open_bag(os.path.join(dir_tmp, name))):
            print("DOI previously resolved. Next file...")
            shutil.rmtree(dir_tmp)

        # Process file if flag does not exist
        else:
            # dir: dir_root -> dir_tmp
            process_lpd(name, dir_tmp)
            # dir: dir_tmp -> dir_root
            os.chdir(dir_root)
            # Zip the directory containing the updated files. Created in dir_root directory
            re_zip(dir_tmp, name, name_ext)
            os.rename(name_ext + '.zip', name_ext)
            # Cleanup and remove tmp directory
            shutil.rmtree(dir_tmp)
    print("NOTE: Quarantine.txt contains a list of errors that may have occurred during processing.")
    print("Process Complete")

    return


def process_lpd(name, dir_tmp):
    """
    Opens up json file, invokes doi_resolver, closes file, updates changelog, cleans directory, and makes new bag.
    :param name: (str) Name of current .lpd file
    :param dir_tmp: (str) Path to tmp directory
    :return: none
    """

    dir_root = os.getcwd()
    dir_bag = os.path.join(dir_tmp, name)
    dir_data = os.path.join(dir_bag, 'data')
    jld_data = {}
    valid = True

    # Navigate down to jLD file
    # dir : dir_root -> dir_data
    os.chdir(dir_data)

    # Open jld file and read in the contents. Execute DOI Resolver.
    with open(os.path.join(dir_data, name + '.jsonld'), 'r') as f:
        try:
            jld_data = json.load(f)
        except ValueError:
            valid = False
            txt_log(dir_root,  name,'quarantine.txt', "Invalid characters. Unable to load file.")

    if valid:
        # Overwrite data with new data
        jld_data = DOIResolver(dir_root, name, jld_data).main()
        # Open the jld file and overwrite the contents with the new data.
        write_json_to_file(os.path.join(dir_data, name + '.jsonld'), jld_data)

        # Open changelog. timestamp it. Prompt user for short description of changes. Close and save
        update_changelog()

    # Delete old bag files, and move files to bag root for re-bagging
    # dir : dir_data -> dir_bag
    dir_cleanup(dir_bag, dir_data)

    # Create a bag for the 3 files
    new_bag = create_bag(dir_bag)
    open_bag(dir_bag)
    new_bag.save(manifests=True)

    return





import io
import os
import urllib.request

import httplib2
import oauth2client
from apiclient import discovery
from oauth2client import client
from oauth2client import tools

from .MapFrame import *


# GLOBALS
APPLICATION_NAME = 'lipd'
SCOPES = 'https://www.googleapis.com/auth/drive'
CLIENT_SECRET_FILE = 'client_secret.json'


def get_google_csv(file_id, filename):
    """
    Get a specific spreadsheet file from Google Drive using its FILE_ID.
    Write the spreadsheet on the local system as a CSV.
    TSNames ID: 1C135kP-SRRGO331v9d8fqJfa3ydmkG2QQ5tiXEHj5us
    :param file_id: (str) Google File ID of target file
    :param filename: (str) Optional: Override filename from google your specified filename
    :return: (str) CSV Filename
    """
    # Download link format
    link_csv = 'https://docs.google.com/spreadsheet/ccc?key=' + file_id + '&output=csv&pref=2'

    # Get authorization and credentials
    credentials = get_credentials()
    http = credentials.authorize(httplib2.Http())

    # A service object represents one google drive account
    service = discovery.build('drive', 'v3', http=http)

    # print("downloading {0}".format(file_id))
    resp, content = service._http.request(link_csv)
    if resp.status == 200:
        # Regex the filename from the response object string
        if not filename:
            for i in resp['content-disposition'].split(';'):
                if 'filename=' in i:
                    try:
                        filename = re.findall(r'"([^"]*)"', i)[0]
                    except IndexError:
                        filename = file_id + '.csv'
                        print("No filename to use. Using File ID.")

        with open(filename, 'wb+') as f:
            f.write(content)
    else:
        print('Error downloading: {0}'.format(file_id))
    return filename


def get_credentials():
    """
    Gets valid user credentials from storage.
    If nothing has been stored, or if the stored credentials are invalid,
    the OAuth2 flow is completed to obtain the new credentials.

    :return: Credentials, the obtained credential.
    """
    home_dir = os.path.expanduser('~')
    credential_dir = os.path.join(home_dir, '.credentials')
    if not os.path.exists(credential_dir):
        os.makedirs(credential_dir)
    credential_path = os.path.join(credential_dir,
                                   CLIENT_SECRET_FILE)

    store = oauth2client.file.Storage(credential_path)
    credentials = store.get()
    if not credentials or credentials.invalid:
        flow = client.flow_from_clientsecrets(CLIENT_SECRET_FILE, SCOPES)
        flow.user_agent = APPLICATION_NAME
        print('Storing credentials to ' + credential_path)
    return credentials


def get_static_google_map(filename_wo_ext, center=None, zoom=None, imgsize=(640,640), imgformat="jpeg",
                          maptype="hybrid", markers=None):
    """
    Retrieve a map (image) from the static google maps server

    Source: http://code.google.com/apis/maps/documentation/staticmaps/
    Source: http://hci574.blogspot.com/2010/04/using-google-maps-static-images.html

    Creates a request string with a URL, looks like:
    http://maps.google.com/maps/api/staticmap?center=Brooklyn+Bridge,New+York,NY&zoom=14&size=512x512&maptype=roadmap
    &markers=color:blue|label:S|40.702147,-74.015794&sensor=false

    :param filename_wo_ext: (str) Filename without extension
    :param center: (tuple of ints) Location to center the map frame
    :param zoom: (int) Zoom amount from 0 (world) to 22 (streets)
    :param imgsize: (tuple of ints) Map image size (max. 640x640)
    :param imgformat: (str) Image format (jpg, bmp, png)
    :param maptype: (str) Map Type (roadmap, satellite, hybrid, terrain)
    :param markers: (list of str) Strings holding marker attributes

    """

    # Assemble the base URL. Append parameters separated by &
    request = "http://maps.google.com/maps/api/staticmap?"

    # If center and zoom  are not given, the map will show all marker locations
    if center is not None:
        request += "center=%s&" % center
    if center is not None:
        request += "zoom=%i&" % zoom

    request += "size=%ix%i&" % imgsize
    request += "format=%s&" % imgformat
    request += "maptype=%s&" % maptype

    # Add markers (location and style)
    if markers is not None:
        for marker in markers:
                request += "%s&" % marker

    # optional: mobile=true assumes the image is shown on a small screen
    # request += "mobile=false&"

    # required, deals with getting location from mobile device
    request += "sensor=false&"
    # print(request)

    # Option 1: save image directly to disk
    # urllib.request.urlretrieve(request, "testing_map."+imgformat)

    # Option 2: read into PIL
    web_sock = urllib.request.urlopen(request)
    img_data = io.BytesIO(web_sock.read())  # constructs a StringIO holding the image

    try:
        # save image to disk, but then open it in Tkinter dialog box
        # if center is None:
        #     center = ''
        pil_img = Image.open(img_data)
        # Display image saved on disk
        # Show image in default image viewer
        pil_img.show()
        # Save as jpg
        pil_img.save(filename_wo_ext+".jpg", "JPEG")
        # m = Tk()
        # m.frame = MapFrame(m, pil_img, str(center))
        # m.mainloop()

    # if this cannot be read as image, it's probably an error from the server,
    except IOError:
        print("IOError:" + str(img_data.read()))  # print error (or it may return a image showing the error"

